/* DO NOT EDIT THIS FILE - it is machine generated */
#include <jni.h>
/* Header for class java_awt_BDGraphics */

#ifndef _WIN32
#undef JNIEXPORT
#define JNIEXPORT static
#endif

#ifndef _Included_java_awt_BDGraphics
#define _Included_java_awt_BDGraphics
#ifdef __cplusplus
extern "C" {
#endif
/*
 * Class:     java_awt_BDGraphics
 * Method:    drawStringN
 * Signature: (JLjava/lang/String;III)V
 */
JNIEXPORT void JNICALL Java_java_awt_BDGraphics_drawStringN
  (JNIEnv *, jobject, jlong, jstring, jint, jint, jint);

#ifdef __cplusplus
}
#endif
#endif
